;nyquist plug-in
;version 3
;type process
;categories "http://lv2plug.in/ns/lv2core#FunctionPlugin"
;name "Find speech segments or beats..."
;action "Calculating..."

;control control-action-choice "Action\n(If not Labels, duplicate your sound\nand then compare the result)" choice "Labels,Boundaries,Graph" 0
;control control-examine-choice "Discard power spectrum above this fraction of sampling rate\n(saves time)" choice "1/2,1/4,1/8,1/16,1/32" 0
;control control-percentile "Find this percentile of remaining spectrum" real "" 50 0 100
;control control-window-length-choice "FFT window Length (samples)" choice "8,16,32,64,128,256,512,1024,2048,4096,8192,16384,32768" 7
;control control-skip-length "Skip Length (samples)" int "" 32 1 32768
;control control-window-choice "Window Type" choice "Hann,Rectangular" 0
;control control-smoothing-window "Smoothing window length (samples)\nNo effect if less than two skips" int "" 0 0 32768
;control control-n-derivatives "Derivatives of log frequency\n(for Boundaries and Labels)" choice "1,2" 0
;control control-sensitivity-octaves "Threshold (+/- octaves/sec or octaves/sec^2)\n(for Boundaries and Labels)" real "" 1400 0 1500000
;control control-n-levels "Show this number of multiples of sensitivity level\n(for Boundaries)" int "" 10 1 100
;control control-minimum-length-ms "Minimum length of labels (ms)" real "" 10 0 1000
;control control-scale "Vertical Scale (Hz)\n(for Graph)" real "" 10000 0 22050
;info 

;(load "Zero Crossings.lsp")
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Find zero crossings, without writing inner loops in Lisp.
;;;
;;; snd is a sound
;;;
;;; choice is 0 for a random-access nearest-crossing finder function that
;;; takes a global time and returns global time or nil;
;;; 1 for an iterator function taking no arguments that returns
;;; (with successive calls) an increasing sequence of global times and then nil.
;;;
;;; Boundary behavior:
;;; The time of the first sample, if that is below tiny and not negative,
;;; and right of the last sample, if that is above -tiny and not positive,
;;; will be reported as a crossing; pass -1 to tiny to suppress that.
(defun make-upward-zero-crossing-finder (snd
					 choice
					 &optional (tiny (db-to-linear -55)))  
  (let*
      ((t0 (snd-t0 snd))
       (srate (snd-srate snd))
       (sample-time (/ (* 2 srate)))
       (blip (snd-from-array 0 srate (vector srate)))
       (blips (trigger snd blip))
       ;; trigger finds transitions from non-positive to
       ;; positive.  But whenever the first sample
       ;; is positive, it is also treated as a crossing.  Reject this
       ;; spurious crossing unless the first sample is "tiny."
       (bogus-crossing-at-t0 (>= (snd-sref snd t0) tiny))
       ;; beware, integrate always puts 0 in the first sample of the result,
       ;; and neglects the last sample of the input; unless input is empty,
       ;; which makes the result empty too.
       (count-of-crossings-snd (integrate blips))
       ;; count-of-crossings array contains what should be
       ;; integers, computed as floats -- be careful to round off.
       ;; Because of the initial 0 always made by integrate,
       ;; (aref count-of-crossings-array index)
       ;; is the count of upward zero transitions BEFORE
       ;; index, not AT OR BEFORE.
       (count-of-crossings-array (snd-samples count-of-crossings-snd ny:all))
       (count-of-crossings-len (length count-of-crossings-array))
       (last-count
	(and (plusp count-of-crossings-len)
	     (aref count-of-crossings-array (1- count-of-crossings-len))))
       ;; Should we report a crossing at or past the end? Find a time or nil.
       ;; This also compensates for the case of loss of the last sample
       ;; to integrate.  The last sample would change the summation if and
       ;; only if there is a transition AT (not past) the last sample,
       ;; which we can detect here.
       ;; If not, the highest value of integrate is attained
       ;; before the last sample and does get seen by snd-inverse;
       ;; though possibly ignored by it, which is compensated for
       ;; elsewhere.
       (extra-crossing
	(let ((s-len (snd-length snd ny:all)))
	  (and (plusp s-len)
	       (let* ((tn (+ t0 (/ (1- s-len) srate)))
		      (sample (snd-sref snd tn)))
		 (if (plusp sample)
		     (and (> s-len 1)
			  (not (plusp
				(snd-sref snd (- tn sample-time))))
			  ;; a rising transition at the very last
			  ;; sample, which is non-negative
			  tn)
		     (and (> sample (- tiny))
			  ;; the last sample is slightly negative, so
			  ;; call it an upward crossing just past the end.
			  (+ tn sample-time))))))))
    (if
     (or (not last-count)		;empty sound
	 (and (not extra-crossing)
	      (or (= last-count 0)	;no real crossings
		  (and bogus-crossing-at-t0 (= last-count 1)))))
					;only bogus crossing
     ;; there are no crossings
     (case choice (0 #'(lambda (time) nil)) (1 #'(lambda () nil)))
     ;; more work to define the function...
     (let*
	 ((inv-snd (snd-inverse count-of-crossings-snd 0 1))
	  ;; notes about snd-inverse:  when the given function is not
	  ;; strictly increasing, the inverse is underdetermined.  Behavior
	  ;; of snd-inverse appears to be that the time of the LAST sample
	  ;; for each value the function attains is the answer given
	  ;; by snd-inverse.  I rely on this assumption!
	  ;; Also snd-inverse seems SOMETIMES to neglect the last sample
	  ;; of input, and sometimes not.
	  ;; Therefore:  take (aref count-of-crossings-array index), then
	  ;; aref inv-array at that value (if it exists), and you get the
	  ;; time of the first upward transition AT OR AFTER index.
	  (inv-array (snd-samples inv-snd ny:all)))

       ;; find the nearest crossing to either side
       ;; or return nil if out of bounds
       (case
	choice
	(0
	 ;; random access nearest-crossing finder
	 #'(lambda (time)
	     (let* ((tt (- time t0))
		    (index0 (truncate (+ 0.5 (* tt srate)))))
	       ;; bounds check for valid time interval
	       (and
		(>= index0 0)
		(< index0 count-of-crossings-len)
		(let*
		    ( ;; Find the time of the first transition AFTER
		     ;; index0 (which is at or after the next index).
		     ;; Or, nil.
		     (index (if (= index0 (1- count-of-crossings-len))
				;; oops, we bump against the problem
				;; of the lost sample in integrate.
				;; But we compensate with
				;; extra-crossing.
				index0
				(1+ index0)))
		     (num (truncate
			   (+ 0.5 (aref count-of-crossings-array index))))
		     ;; in case num equals last-count, again rely on
		     ;; extra-crossing.
		     (t-after (if (< num last-count)
				  (aref inv-array num)
				  extra-crossing))
		     ;; Are there any real crossings at or before
		     ;; index?
		     (should-look-left
		      (not (or (zerop num)
			       (and (= num 1) bogus-crossing-at-t0))))
		     ;; Time of the first real transition
		     ;; AT OR BEFORE index 0, or nil:
		     (t-before
		      (and should-look-left
			   (aref inv-array (1- num)))))
		  (if (and t-before
			   (or (not t-after)
			       (< (- time t-before) (- t-after time))))
		      t-before
		      t-after))))))
	(1
	 ;; iterator, a function with state
	 (let ((num (if bogus-crossing-at-t0 1 0))
	       (last (length inv-array))
	       (extra extra-crossing))
	   #'(lambda ()
	       (cond ((< num last)
		      (prog1 (aref inv-array num)
			(setq num (1+ num))))
		     ((numberp extra)
		      (prog1 extra (setq extra nil)))
		     (t nil))))))))))
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; take the array of fft coefficients
;;; use nyquist to take advantage of faster array operations (on sound objects)
;;; than we can write in Lisp
(defun find-multiplier (frame spectrum-length-divisor threshold-fraction)
  ;; first make the dc component comparable to the other components
  (setf (aref frame 0) (* 0.5 (aref frame 0)))
  ;; (prog1
  (let* ((len (/ (length frame) spectrum-length-divisor))

	 (frame-snd (snd-from-array 0.0 1.0 frame))    ;array as sound
	 (frame-sq-snd (mult frame-snd frame-snd))     ;the squares

	 ;; following finds partial sums, always gives 0 in first
	 ;; place, and ignores the last component of frame-sq-snd which is
	 ;; the Nyquist frequency.  I would divide it by 2 first as with DC
	 ;; and pad with a zero if I cared about it.
	 (power-spectrum-snd (integrate frame-sq-snd)))

    (let*
	((spectrum (snd-fetch-array power-spectrum-snd len 1))
	 (last (aref spectrum (1- len)))
	 (threshold (* last (min 1.0 threshold-fraction))))

      ;; Now search odd number indices only (because they sum sine
      ;; and cosine components in pairs)

      ;; ;; sref-inverse version
      ;; ;; does not give equivalent results
      ;; (let
      ;; 	  ((jj (/ (truncate
      ;; 		   (sref-inverse power-spectrum-snd threshold)) 2)))
      ;; 	(do ((ii (max 1 (1- jj)) (+ 2 ii)))
      ;; 	    ((or (>= ii len)
      ;; 		 (>= (aref spectrum ii) threshold)) (/ ii 2))))

      ;; linear search version
      ;; assuming length is even, put a sentinel in place to simplify the
      ;; loop test
      (setf (aref spectrum (1- len)) threshold)
      (do ((ii 1 (+ 2 ii)))
      	  ((>= (aref spectrum ii) threshold) (/ ii 2)))

      ;; ;; bisection search
      ;; ;; doesn't seem to improve time
      ;; ;; this relies on len being a power of 2
      ;; ;; invariants:  low is even; the desired array element is after
      ;; ;; low and at or before low + 2 * diff; diff is a power of 2
      ;; (do ((low -1)
      ;; 	   (diff (/ len 2) (/ diff 2)))
      ;; 	  ((= 1 diff) (/ (+ low 2) 2))
      ;; 	(if (< (aref spectrum (+ low diff)) threshold)
      ;; 	    (setf low (+ low diff))))

      ))

  ;; should be nice and leave no net side effect.
  ;; but trying to shave every bit of time.
  ;; (setf (aref frame 0) (* 2 (aref frame 0)))
  ;; )					; prog1
  )

;;; last arg is a function taking a frame
(defun scan-frames (snd length-samples skip-samples window frame-visitor)
  (let ((my-s (snd-copy snd)))
    (do ((frame (snd-fft my-s length-samples skip-samples window)
		(snd-fft my-s length-samples skip-samples window)))
	((not (arrayp frame)))
      (funcall frame-visitor frame))))

(defun hann-window (window-length-seconds)
  (let ((window-freq-hz (/ window-length-seconds)))
    (mult 0.5
	  (sum 1 (osc (hz-to-step window-freq-hz)
		      (/ window-length-seconds (get-duration 1))
		      *sine-table* -90)))))

(defun make-graph (snd
		   spectrum-length-divisor
		   threshold-fraction
		   window-length
		   skip-length
		   window-choice)
  (let* ((srate (snd-srate snd))
	 (reduced-srate (/ srate skip-length))
	 (snd-len (snd-length snd ny:all))
	 ;;correct graph length for the final frames that extend beyond the
	 ;;sound data:  don't include those.
	 (graph (make-array (/ (- snd-len (/ window-length 2)) skip-length)))
	 (window-length-seconds (/ window-length srate))
	 (window-freq-hz (/ window-length-seconds))
	 (window
	  (case window-choice
		(0 (hann-window window-length-seconds))
		(1 nil)))
	 (nn 0)
	 (visit-frame
	  #'(lambda (frame)
	      (if (< nn (length graph))
		  (let ((value (max 0.1	;guard against taking log 0 later
				    (* (find-multiplier
					frame spectrum-length-divisor
					threshold-fraction)
				       window-freq-hz))))
		    (setf (aref graph nn) value)))
	      (setq nn (1+ nn)))))
    
    ;;compute rest of graph
    (scan-frames snd window-length skip-length window visit-frame)

    ;; convert array to a sound whose sample rate is divided by the skip
    (snd-from-array (snd-t0 snd) reduced-srate graph)))

(defun smoothing-function (srate smoothing-window)
  (let* ((window-seconds (/ smoothing-window srate))
	 ;; Use the Hann window, scaled to integrate to 1,
	 ;; why not?  Experiment with others later.
	 ;; Nyquist's integrate function divides by the sample rate, but
	 ;; convolve does not, so I do that here.
	 (hw (sound-srate-abs srate (hann-window window-seconds))))
    (prod (* 2.0 (/ window-seconds) (/ srate)) hw)))

(defun smoothed-graph (g-snd smoothing-window)
  (convolve g-snd
	    (smoothing-function (snd-srate g-snd) smoothing-window)))

(defun smoothed-derivative-of-log (graph-snd n-derivatives smoothing-window)
  ;; derivative of a convolution equals the convolution of the derivative
  (let* ((log-graph-snd (s-log graph-snd))
	 (snd (case n-derivatives
		    (0 log-graph-snd)
		    (1 (slope log-graph-snd))
		    (2 (slope (slope log-graph-snd))))))
    (if (< smoothing-window 2)
	snd
	(smoothed-graph snd smoothing-window))))

(defun find-correction (snd window-length skip-length
			    n-derivatives smoothing-window)
  (let* ((srate (snd-srate snd))
	 (reduced-srate (/ srate skip-length)))
    (+ (* window-length 0.5 (/ srate))	; for fft
       ;; Each call of slope shifted the results a half sample earlier
       (* n-derivatives 0.5 (/ reduced-srate))
       ;; Any post-smoothing shifted results half a window
       (if (< smoothing-window 2)
	   0
	   (* smoothing-window -0.5 (/ reduced-srate))))))

(defun find-cliffs (deriv-snd skip-length n-levels sensitivity-octaves)
  (let* ((first-trigger-level (* (log 2.0) sensitivity-octaves))
	 (srate (snd-srate deriv-snd))
	 (single-positive (snd-from-array 0 srate (vector (/ 1.0 n-levels))))
	 (single-negative (prod -1 single-positive))
	 (result (s-rest 0)))
    (dotimes (ii n-levels)
      (let* ((trigger-level (* (1+ ii) first-trigger-level))
	     (sig-rising (diff deriv-snd trigger-level))
	     (sig-falling (diff (prod -1 deriv-snd) trigger-level)))
	(setq result (sum result
			  (trigger sig-rising single-positive)
			  (trigger sig-falling single-negative)))))
    result))

(defun merge-close-times (iter minimum)
  (let* ((last-time (funcall iter)))	; lookahead
    #'(lambda ()
	;; average times until that differs enough from the lookahead time.
	(and last-time
	     (do*
		 ((sum last-time)
		  (nn 1 (1+ nn))
		  (tt (funcall iter) (funcall iter)))
		 ((or (not tt) (>= (- tt (/ sum nn)) minimum))
		  (setq last-time tt)
		  (/ sum nn))
	       (setq sum (+ sum tt)))))))

(defun map-iter (f iter)
  #'(lambda ()
      (let ((time (funcall iter)))
	(and time (funcall f time)))))

(defun find-labels (snd deriv-snd correction minimum-length-seconds
			window-length sensitivity-octaves)
  (let* ((trigger-level (* (log 2.0) sensitivity-octaves))
	 (trigger-snd (diff (s-abs deriv-snd) trigger-level))
	 (iter0 (merge-close-times
		 (make-upward-zero-crossing-finder trigger-snd 1)
		 minimum-length-seconds))
	 (refiner (make-upward-zero-crossing-finder snd 0))
	 (iter (map-iter
		#'(lambda (time) (funcall refiner (+ time correction)))
		iter0))
	 (prev-time (funcall iter))
	 results)
    (do* ((time (funcall iter) (funcall iter)))
	((not time) results)
      (setq results (cons (list prev-time time "") results)
	    prev-time time))))

(defun scale-horz-for-display (graph-snd correction snd)
  ;; must correct the starting time, and interpolate for increased
  ;;sample rate
  (let* ((srate (snd-srate snd))
	 (scaled-snd (force-srate srate graph-snd))
	 (local-correction-time
	  (/ (* correction srate) (snd-length snd ny:all))))
    (if (plusp correction)
	(extract 0.0 1.0   ; correct for samples that convolve adds to the end
		 (sum (s-rest (* 2 correction))
		      (at local-correction-time  (cue scaled-snd))))
	(extract local-correction-time 1.0 scaled-snd))))

(defun power-of-2 (ii)
  (let ((result 1))
    (dotimes (_ ii)
      (setf result (* 2 result)))
    result))

;; convert any stereo input to the average of channels
(defun mono-s (snd)
  (if (arrayp snd)
      (prod 0.5 (sum (aref snd 0) (aref snd 1)))
      snd))

;;; top level invocations, I like to restrict use of global control variables
;;; just to here
(let*
    ((snd (mono-s s))
     (window-length (power-of-2 (+ 3 control-window-length-choice)))
     (just-graphing (= control-action-choice 2))
     (n-derivatives
      (if just-graphing
	  0
	  (1+ control-n-derivatives)))
     (spectrum-length-divisor (power-of-2 control-examine-choice))
     (threshold-fraction (/ control-percentile 100.0))
     (minimum-length-seconds (/ control-minimum-length-ms 1000.0))
     ;; The user is queried for a smoothing window in terms of the
     ;; input sound, but it gets applied a downsampled graph
     ;; "sound."
     (smoothing-window (/ control-smoothing-window control-skip-length))
     (correction (find-correction snd window-length control-skip-length
				  n-derivatives
				  smoothing-window))
     (graph-snd
      (make-graph snd
		  spectrum-length-divisor
		  threshold-fraction
		  window-length
		  control-skip-length
		  control-window-choice))
     (result
      (if just-graphing
	  (let ((intermediate-graph
		 (if (< smoothing-window 2)
		     graph-snd
		     ;; apply convolution to the log, not the graph itself
		     (s-exp (smoothed-derivative-of-log
			     graph-snd 0 smoothing-window)))))
	    (clip (mult (/ control-scale) intermediate-graph) 1.0))
	  (let ((deriv-snd (smoothed-derivative-of-log
			    graph-snd
			    n-derivatives
			    smoothing-window)))
	    (case control-action-choice
		  (0 (find-labels snd
				  deriv-snd
				  correction
				  minimum-length-seconds
				  window-length
				  control-sensitivity-octaves))
		  (1 (find-cliffs deriv-snd
				  control-skip-length
				  control-n-levels
				  control-sensitivity-octaves)))))))
  (if (soundp result)
      ;; Fit horizontally and replicate in both channels if stereo
      (let ((scaled-result
	     (scale-horz-for-display result correction snd)))
	(if (arrayp s)
	    (vector scaled-result scaled-result)
	    scaled-result))
      ;; list of labels, maybe empty
      (or result "Found no boundaries")))
